ANY = "Any"
BCSVF = "bimcollabsmartviewfile"
VERSION = "version"
APPVER = "applicationversion"
SMVSETS = "SMARTVIEWSETS"
SMVSET = "SMARTVIEWSET"
TITLE = "TITLE"
DESCRIPTION = "DESCRIPTION"
GUID = "GUID"
MODDATE = "MODIFICATIONDATE"
SVIEWS = "SMARTVIEWS"
SVIEW = "SMARTVIEW"
CREATOR = "CREATOR"
CREATIONDATE = "CREATIONDATE"
MODIFIER = "MODIFIER"
MODIFICATIONDATE = "MODIFICATIONDATE"
RULES = "RULES"
RULE = "RULE"
ITAKEOFF = "INFORMATIONTAKEOFF"
ITAOPSET = "NFORMATIONTAKEOFF.PROPERTYSETNAME"
ITAOPN = "INFORMATIONTAKEOFF.PROPERTYNAME"
ITAOPER = "INFORMATIONTAKEOFF.OPERATION"
IFCTYPE = "IFCTYPE"
NAME = "NAME"
PSETNAME = "PROPERTYSETNAME"
TYPE = "TYPE"
VALUETYPE = "VALUETYPE"
UNIT = "UNIT"
CONDITION = "CONDITION"
STATE = "STATE"
VALUE = "VALUE"
PROPERTY = "PROPERTY"
PROPERTYSET = "PropertySet"
NONE = "None"



from __future__ import annotations
import SOMcreator
from SOMcreator import classes


def _find_parent(uuid_dict, element: classes.ClassTypes):
    for test_el, identifier in SOMcreator.filehandling.parent_dict.items():
        if type(test_el) is not type(element):
            continue
        if identifier not in uuid_dict:
            continue
        if test_el == element:
            continue
        if test_el.parent is not None:
            continue

        if test_el.name != element.name:
            continue

        if isinstance(test_el, classes.Attribute):
            if test_el.value == element.value:
                return identifier
        return identifier


def calculate():
    uuid_dict = classes.get_uuid_dict()
    for entity, uuid in SOMcreator.filehandling.parent_dict.items():
        if uuid is None:
            continue
        if uuid not in uuid_dict:
            uuid = _find_parent(uuid_dict, entity)
        if uuid is None:
            continue
        uuid_dict[uuid].add_child(entity)



from __future__ import annotations

import os
from datetime import date

from lxml.etree import Element, ElementTree, SubElement

from . import ids_xsd, xml_xsd
from ... import classes
from ...constants import value_constants, ifc_datatypes

NSMAP = {None: ids_xsd.DEFAULT_NS[1:-1],
         "xs": xml_xsd.NS_XS[1:-1],
         "xsi": xml_xsd.NS_XSI[1:-1]}


def _build_info(proj: classes.Project, author, xml_parent: Element) -> None:
    xml_element = SubElement(xml_parent, ids_xsd.INFO, nsmap=NSMAP)
    SubElement(xml_element, ids_xsd.TITLE, nsmap=NSMAP).text = f"Pruefregeln fuer Projekt '{proj.name}'"
    SubElement(xml_element, ids_xsd.COPYRIGHT, nsmap=NSMAP).text = "None"
    SubElement(xml_element, ids_xsd.VERSION, nsmap=NSMAP).text = f"0.9"
    SubElement(xml_element, ids_xsd.DESCRIPTION, nsmap=NSMAP).text = f"Autogenerated by SOMcreator"
    SubElement(xml_element, ids_xsd.AUTHOR, nsmap=NSMAP).text = str(author)
    SubElement(xml_element, ids_xsd.DATE, nsmap=NSMAP).text = f"{date.today()}"
    SubElement(xml_element, ids_xsd.PURPOSE, nsmap=NSMAP).text = "Modelcheck"


def _build_specifications(required_data, xml_parent: Element) -> None:
    xml_specifications = SubElement(xml_parent, ids_xsd.SPECIFICATIONS, nsmap=NSMAP)

    for obj, property_set_dict in required_data.items():
        if obj.ident_attrib is None:
            continue
        _build_specification(obj, property_set_dict, xml_specifications)


def _build_applicability(obj: classes.Object, xml_parent: Element) -> None:
    xml_applicability = SubElement(xml_parent, ids_xsd.APPLICABILITY, nsmap=NSMAP)
    xml_property = SubElement(xml_applicability, ids_xsd.PROPERTY, nsmap=NSMAP)
    xml_property.set(ids_xsd.ATTR_DATATYPE, ifc_datatypes.LABEL)
    xml_property_set = SubElement(xml_property, ids_xsd.PROPERTYSET, nsmap=NSMAP)
    SubElement(xml_property_set, ids_xsd.SIMPLEVALUE).text = obj.ident_attrib.property_set.name
    xml_name = SubElement(xml_property, ids_xsd.NAME, nsmap=NSMAP)
    SubElement(xml_name, ids_xsd.SIMPLEVALUE, nsmap=NSMAP).text = obj.ident_attrib.name
    xml_value = SubElement(xml_property, ids_xsd.VALUE)
    SubElement(xml_value, ids_xsd.SIMPLEVALUE, nsmap=NSMAP).text = obj.ident_value


def _build_requirements(property_set_dict: dict[classes.PropertySet, list[classes.Attribute]],
                        xml_parent: Element) -> None:
    xml_requirement = SubElement(xml_parent, ids_xsd.REQUIREMENTS, nsmap=NSMAP)
    for property_set, attribute_list in property_set_dict.items():
        for attribute in attribute_list:
            _build_attribute_requirement(attribute, xml_requirement)


def _build_attribute_requirement(attribute: classes.Attribute, xml_parent: Element) -> None:
    xml_property = SubElement(xml_parent, ids_xsd.PROPERTY, nsmap=NSMAP)
    xml_property.set(ids_xsd.ATTR_DATATYPE, attribute.data_type)
    if attribute.optional:
        xml_property.set(xml_xsd.MINOCCURS, "0")
    else:
        xml_property.set(xml_xsd.MINOCCURS, "1")
    xml_property.set(xml_xsd.MAXOCCURS, "1")

    xml_property_set = SubElement(xml_property, ids_xsd.PROPERTYSET)
    SubElement(xml_property_set, ids_xsd.SIMPLEVALUE, nsmap=NSMAP).text = attribute.property_set.name
    xml_name = SubElement(xml_property, ids_xsd.NAME)
    SubElement(xml_name, ids_xsd.SIMPLEVALUE, nsmap=NSMAP).text = attribute.name
    if not attribute.value:
        return
    xml_value = SubElement(xml_property, ids_xsd.VALUE, nsmap=NSMAP)
    xml_restriction = SubElement(xml_value, xml_xsd.RESTRICTION, nsmap=NSMAP)
    xml_restriction.set(xml_xsd.BASE, value_constants.XS_DATATYPE_DICT[attribute.data_type])

    if attribute.value_type == value_constants.LIST:
        for value in attribute.value:
            SubElement(xml_restriction, xml_xsd.ENUMERATION, nsmap=NSMAP).set(xml_xsd.VALUE, str(value))

    if attribute.value_type == value_constants.RANGE:
        min_value = min(min(v[0] for v in attribute.value), min(v[1] for v in attribute.value))
        max_value = max(max(v[0] for v in attribute.value), max(v[1] for v in attribute.value))
        SubElement(xml_restriction, xml_xsd.MININCLUSIVE, nsmap=NSMAP).set(xml_xsd.VALUE, str(min_value))
        SubElement(xml_restriction, xml_xsd.MAXINCLUSIVE, nsmap=NSMAP).set(xml_xsd.VALUE, str(max_value))

    if attribute.value_type == value_constants.FORMAT:
        pattern = "|".join(attribute.value)
        SubElement(xml_restriction, xml_xsd.PATTERN, nsmap=NSMAP).set(xml_xsd.VALUE, pattern)


def _build_specification(obj: classes.Object, property_set_dict: dict[classes.PropertySet, list[classes.Attribute]],
                         xml_parent: Element) -> None:
    xml_specification = SubElement(xml_parent, ids_xsd.SPECIFICATION, nsmap=NSMAP)
    xml_specification.set(ids_xsd.ATTR_NAME, f"Pruefregel {obj.name} ({obj.ident_value})")
    xml_specification.set(ids_xsd.ATTR_IFCVERSION, ids_xsd.VAL_IFC4)
    xml_specification.set(ids_xsd.ATTR_DESCRIPTION, "Automatisch generierte Attributpruefregel")
    xml_specification.set(xml_xsd.MINOCCURS, "0")
    xml_specification.set(xml_xsd.MAXOCCURS, "unbounded")
    _build_applicability(obj, xml_specification)
    _build_requirements(property_set_dict, xml_specification)


def export(proj: classes.Project,
           required_data: dict[classes.Object, dict[classes.PropertySet, list[classes.Attribute]]],
           path: str | os.PathLike, author=None) -> None:
    if not author:
        author = proj.author
    xml_root = Element(ids_xsd.IDS, nsmap=NSMAP)
    xml_root.set(xml_xsd.NS_XSI + xml_xsd.SCHEMALOCATION, ids_xsd.SCHEME_LOCATION_NS)
    _build_info(proj, author, xml_root)
    _build_specifications(required_data, xml_root)
    ElementTree(xml_root).write(path, pretty_print=True, encoding="UTF-8")



NS_XSI = "{http://www.w3.org/2001/XMLSchema-instance}"
NS_XS = "{http://www.w3.org/2001/XMLSchema}"
# Classes
ALL = NS_XS + 'all'
ANNOTATED = NS_XS + 'annotated'
ANNOTATION = NS_XS + 'annotation'
ANY = NS_XS + 'any'
ANYATTRIBUTE = NS_XS + 'anyAttribute'
ANYTYPE = NS_XS + 'anyType'
APPINFO = NS_XS + 'appinfo'
ATTRIBUTE = NS_XS + 'attribute'
ATTRIBUTEGROUP = NS_XS + 'attributeGroup'
ATTRIBUTEGROUPREF = NS_XS + 'attributeGroupRef'
CHOICE = NS_XS + 'choice'
COMPLEXCONTENT = NS_XS + 'complexContent'
COMPLEXRESTRICTIONTYPE = NS_XS + 'complexRestrictionType'
COMPLEXTYPE = NS_XS + 'complexType'
DOCUMENTATION = NS_XS + 'documentation'
ELEMENT = NS_XS + 'element'
ENUMERATION = NS_XS + 'enumeration'
EXPLICITGROUP = NS_XS + 'explicitGroup'
EXTENSION = NS_XS + 'extension'
EXTENSIONTYPE = NS_XS + 'extensionType'
FACET = NS_XS + 'facet'
FIELD = NS_XS + 'field'
FRACTIONDIGITS = NS_XS + 'fractionDigits'
GROUP = NS_XS + 'group'
GROUPREF = NS_XS + 'groupRef'
IMPORT = NS_XS + 'import'
INCLUDE = NS_XS + 'include'
KEY = NS_XS + 'key'
KEYBASE = NS_XS + 'keybase'
KEYREF = NS_XS + 'keyref'
LENGTH = NS_XS + 'length'
LIST = NS_XS + 'list'
LOCALCOMPLEXTYPE = NS_XS + 'localComplexType'
LOCALELEMENT = NS_XS + 'localElement'
LOCALSIMPLETYPE = NS_XS + 'localSimpleType'
MAXEXCLUSIVE = NS_XS + 'maxExclusive'
MAXINCLUSIVE = NS_XS + 'maxInclusive'
MAXLENGTH = NS_XS + 'maxLength'
MINEXCLUSIVE = NS_XS + 'minExclusive'
MININCLUSIVE = NS_XS + 'minInclusive'
MINLENGTH = NS_XS + 'minLength'
NAMEDATTRIBUTEGROUP = NS_XS + 'namedAttributeGroup'
NAMEDGROUP = NS_XS + 'namedGroup'
NARROWMAXMIN = NS_XS + 'narrowMaxMin'
NOFIXEDFACET = NS_XS + 'noFixedFacet'
NOTATION = NS_XS + 'notation'
NUMFACET = NS_XS + 'numFacet'
OPENATTRS = NS_XS + 'openAttrs'
PATTERN = NS_XS + 'pattern'
REALGROUP = NS_XS + 'realGroup'
REDEFINE = NS_XS + 'redefine'
RESTRICTION = NS_XS + 'restriction'
RESTRICTIONTYPE = NS_XS + 'restrictionType'
SCHEMA = NS_XS + 'schema'
SELECTOR = NS_XS + 'selector'
SEQUENCE = NS_XS + 'sequence'
SIMPLECONTENT = NS_XS + 'simpleContent'
SIMPLEEXPLICITGROUP = NS_XS + 'simpleExplicitGroup'
SIMPLEEXTENSIONTYPE = NS_XS + 'simpleExtensionType'
SIMPLERESTRICTIONTYPE = NS_XS + 'simpleRestrictionType'
SIMPLETYPE = NS_XS + 'simpleType'
TOPLEVELATTRIBUTE = NS_XS + 'topLevelAttribute'
TOPLEVELCOMPLEXTYPE = NS_XS + 'topLevelComplexType'
TOPLEVELELEMENT = NS_XS + 'topLevelElement'
TOPLEVELSIMPLETYPE = NS_XS + 'topLevelSimpleType'
TOTALDIGITS = NS_XS + 'totalDigits'
UNION = NS_XS + 'union'
UNIQUE = NS_XS + 'unique'
WHITESPACE = NS_XS + 'whiteSpace'
WILDCARD = NS_XS + 'wildcard'

# Attributes
ABSTRACT = 'abstract'
ATTRIBUTEFORMDEFAULT = 'attributeFormDefault'
BASE = 'base'
BLOCK = 'block'
BLOCKDEFAULT = 'blockDefault'
DEFAULT = 'default'
ELEMENTFORMDEFAULT = 'elementFormDefault'
FINAL = 'final'
FINALDEFAULT = 'finalDefault'
FIXED = 'fixed'
FORM = 'form'
ID = 'id'
ITEMTYPE = 'itemType'
MAXOCCURS = 'maxOccurs'
MEMBERTYPES = 'memberTypes'
MINOCCURS = 'minOccurs'
MIXED = 'mixed'
NAME = 'name'
NAMESPACE = 'namespace'
NILLABLE = 'nillable'
PROCESSCONTENTS = 'processContents'
PUBLIC = 'public'
REF = 'ref'
REFER = 'refer'
SCHEMALOCATION = 'schemaLocation'
SOURCE = 'source'
SUBSTITUTIONGROUP = 'substitutionGroup'
SYSTEM = 'system'
TARGETNAMESPACE = 'targetNamespace'
TYPE = 'type'
USE = 'use'
VALUE = 'value'
VERSION = 'version'
XPATH = 'xpath'

# Values
VAL_0 = '0'
VAL_1 = '1'
VAL_COLLAPSE = 'collapse'
VAL_EXTENSION = 'extension'
VAL_LAX = 'lax'
VAL_LIST = 'list'
VAL_OPTIONAL = 'optional'
VAL_PRESERVE = 'preserve'
VAL_PROHIBITED = 'prohibited'
VAL_QUALIFIED = 'qualified'
VAL_REPLACE = 'replace'
VAL_REQUIRED = 'required'
VAL_RESTRICTION = 'restriction'
VAL_SKIP = 'skip'
VAL_STRICT = 'strict'
VAL_SUBSTITUTION = 'substitution'
VAL_UNBOUNDED = 'unbounded'
VAL_UNION = 'union'
VAL_UNQUALIFIED = 'unqualified'




from __future__ import annotations
from typing import TYPE_CHECKING
import logging
from SOMcreator import classes
import SOMcreator
from SOMcreator.filehandling.constants import PROJECT_PHASES, USE_CASES, NAME, DESCRIPTION, OPTIONAL, PARENT, \
    FILTER_MATRIX

if TYPE_CHECKING:
    from SOMcreator import Project
    from SOMcreator.filehandling.typing import ProjectDict, StandardDict, ObjectDict, PropertySetDict, AttributeDict, \
        AggregationDict


##### Import #####

def get_filter_lists(project_dict: ProjectDict):
    project_phases = project_dict.get(PROJECT_PHASES)
    use_cases = project_dict.get(USE_CASES)
    phase_list = list()
    use_case_list = list()
    if project_phases is not None and isinstance(project_phases, list):
        for phase in project_phases:
            if isinstance(phase, str):
                phase_list.append(classes.Phase(phase, phase, ""))
            else:
                phase_list.append(classes.Phase(phase["name"], phase.get("long_name"), phase.get("description")))
    else:
        phase_list = [classes.Phase("Stand", "Standard", "Automatisch generiert bitte aktualisieren")]
    if use_cases is not None and isinstance(use_cases, list):
        for use_case in use_cases:
            if isinstance(use_case, str):
                use_case_list.append(classes.UseCase(use_case, use_case, ""))
            else:
                use_case_list.append(
                    classes.UseCase(use_case["name"], use_case.get("long_name"), use_case.get("description")))
    else:
        use_case_list = [classes.UseCase("Stand", "Standard", "Automatisch generiert bitte aktualisieren")]
    return phase_list, use_case_list


def get_basics(proj: SOMcreator.Project, element_dict: StandardDict) -> tuple[str, str, bool, str, list[list[bool]]]:
    def get_value(d: dict, p: str) -> bool:
        return d.get(p) if d.get(p) is not None else True

    name = element_dict[NAME]
    description = element_dict[DESCRIPTION]
    optional = element_dict[OPTIONAL]
    parent = element_dict[PARENT]
    matrix = element_dict.get(FILTER_MATRIX)
    phase_list, use_case_list = SOMcreator.filehandling.phase_list, SOMcreator.filehandling.use_case_list
    matrix_list = list()
    for _ in proj.get_project_phase_list():
        matrix_list.append([True for __ in proj.get_use_case_list()])

    if matrix is None:  # handle deprecated file types
        matrix = list()
        file_phases: list[bool] = element_dict.get(PROJECT_PHASES)
        if isinstance(file_phases, dict):  # deprecated
            output_phases = [get_value(file_phases, phase.name) for phase in proj.get_project_phase_list()]
        elif file_phases is None:
            output_phases = [True for _ in proj.get_project_phase_list()]
        else:
            output_phases = [True for _ in proj.get_project_phase_list()]
            for output_index, existing_phase in enumerate(proj.get_project_phase_list()):
                if existing_phase in phase_list:
                    value = file_phases[phase_list.index(existing_phase)]
                    output_phases[output_index] = value

        file_use_cases: list[bool] = element_dict.get(USE_CASES)
        if file_use_cases is None:
            output_use_cases = [True for _ in proj.get_use_case_list()]
        else:
            output_use_cases = [True for _ in proj.get_use_case_list()]
            for output_index, existing_use_case in enumerate(proj.get_use_case_list()):
                if existing_use_case in use_case_list:
                    value = file_use_cases[use_case_list.index(existing_use_case)]
                    output_use_cases[output_index] = value

        for phase_index, phase in enumerate(output_phases):
            pl = list()
            for use_case_index, use_case in enumerate(output_use_cases):
                pl.append(bool(output_phases[phase_index] and output_use_cases[use_case_index]))
            matrix.append(pl)
    return name, description, optional, parent, matrix


def check_dict(d: dict | None, d_name: str) -> bool:
    if d is None:
        logging.error(f"loading Error: {d_name} doesn't exist!")
        return True
    return False


def remove_part_of_dict(key):
    """
    Removes part of plugin dict if its saved in Core
    :param key:
    :return:
    """
    SOMcreator.filehandling.plugin_dict.pop(key)


#### Export ######

def write_filter_matrix(element: classes.ClassTypes):
    proj = element.project
    phases = proj.get_project_phase_list()
    use_cases = proj.get_use_case_list()
    matrix = list()
    for phase in phases:
        phase_list = list()
        for use_case in use_cases:
            phase_list.append(element.get_filter_state(phase, use_case))
        matrix.append(phase_list)
    return matrix


def write_basics(entity_dict: ObjectDict | PropertySetDict | AttributeDict | AggregationDict,
                 element: classes.ClassTypes) -> None:
    """function gets called from all Entities"""
    entity_dict[NAME] = element.name
    entity_dict[OPTIONAL] = element.optional
    entity_dict[FILTER_MATRIX] = write_filter_matrix(element)
    parent = None if element.parent is None else element.parent.uuid
    entity_dict[PARENT] = parent
    entity_dict[DESCRIPTION] = element.description



import logging

from openpyxl import load_workbook, Workbook

from .. import classes


def create_mapping(src_path: str, dest_path: str, project: classes.Project) -> None:
    def _create_sheet(obj: classes.Object, workbook: Workbook, name):
        new_sheet = workbook.create_sheet(name)
        attributes = set()
        for property_set in obj.property_sets:
            for attribute in property_set.attributes:
                attributes.add(attribute.name)

        for i, attrib_name in enumerate(sorted(attributes), start=1):
            new_sheet.cell(1, i).value = attrib_name

    export_wb = Workbook()
    export_wb.active.title = "Hilfe"
    wb = load_workbook(src_path)
    sheet = wb.active
    important_rows = [row for i, row in enumerate(sheet.rows) if row[2].value is not None and i != 0]
    object_dict = {obj.ident_attrib.value[0]: obj for obj in project.objects if
                   not obj.is_concept}

    for row in important_rows:
        bauteil_bez_card, bauteil_bez_2, bauteilklass = map(lambda x: x.value, row)
        obj = object_dict.get(bauteilklass)
        if obj is None:
            logging.warning(f"identifier '{bauteilklass}' not found")
            continue
        _create_sheet(obj, export_wb, bauteil_bez_card)
    export_wb.save(dest_path)



from .classes import PropertySet, Object, Project, Attribute
from .external_software import desite, excel, allplan, revit, vestra, card1
from .constants import json_constants, value_constants
from .tools import merge_projects

__version__ = "1.6.11"
active_project = None



DESCRIPTION = "description"
OPTIONAL = "optional"
CURRENT_PR0JECT_PHASE = "current_project_phase"
CURRENT_USE_CASE = "current_use_case"
PROJECT_PHASES = "ProjectPhases"
USE_CASES = "UseCases"
FILTER_MATRIX = "filter_matrix"
AGGREGATION_PSET = "AggregationPsetName"
AGGREGATION_ATTRIBUTE = "AggregationAttributeName"
PREDEFINED_PSETS = "PredefinedPropertySets"
PROPERTY_SETS = "PropertySets"
IDENT_ATTRIBUTE = "ident_attribute"
ATTRIBUTES = "Attributes"
OBJECT = "Object"
OBJECTS = "Objects"
AGGREGATIONS = "Aggregations"
NAME = "name"
PARENT = "parent"
DATA_TYPE = "data_type"
VALUE_TYPE = "value_type"
CHILD_INHERITS_VALUE = "child_inherits_value"
PROJECT = "Project"
VERSION = "version"
AUTHOR = "author"
X_POS = "x_pos"
Y_POS = "y_pos"
CONNECTION = "connection"
IFC_MAPPINGS = "IfcMappings"
IFC_MAPPING = "IfcMapping"
ABBREVIATION = "abbreviation"
REVIT_MAPPING = "revit_mapping"
VALUE = "Value"
IGNORE_PSET = "IFC"


from __future__ import annotations

import os
import json
import SOMcreator
from .typing import MainDict
from typing import Type, TYPE_CHECKING
from . import constants, core, project, predefined_pset, property_set, obj, aggregation, inheritance

if TYPE_CHECKING:
    from SOMcreator.classes import Project
parent_dict = dict()
aggregation_dict = dict()
phase_list = list()
use_case_list = list()
plugin_dict = dict()


def open_json(cls: Type[Project], path: str):
    SOMcreator.filehandling.parent_dict = dict()

    if not os.path.isfile(path):
        return

    with open(path, "r") as file:
        main_dict: MainDict = json.load(file)

    SOMcreator.filehandling.plugin_dict = dict(main_dict)

    project_dict = main_dict.get(constants.PROJECT)
    SOMcreator.filehandling.phase_list, SOMcreator.filehandling.use_case_list = core.get_filter_lists(project_dict)

    proj, project_dict = project.load(cls, main_dict)
    predefined_pset.load(proj, main_dict)

    obj.load(proj, main_dict)

    aggregation.load(proj, main_dict)

    inheritance.calculate()
    aggregation.calculate()
    proj.plugin_dict = SOMcreator.filehandling.plugin_dict
    proj.import_dict = main_dict
    return proj


def export_json(proj: Project, path: str) -> dict:
    main_dict: MainDict = dict()
    project.write(proj, main_dict)
    predefined_pset.write(proj, main_dict)
    obj.write(proj, main_dict)
    aggregation.write(proj, main_dict)

    main_dict.update(proj.plugin_dict)
    with open(path, "w") as file:
        json.dump(main_dict, file)
    return main_dict



ABSORBEDDOSEMEASURE = 'IfcAbsorbedDoseMeasure'
ACCELERATIONMEASURE = 'IfcAccelerationMeasure'
AMOUNTOFSUBSTANCEMEASURE = 'IfcAmountOfSubstanceMeasure'
ANGULARVELOCITYMEASURE = 'IfcAngularVelocityMeasure'
ARCINDEX = 'IfcArcIndex'
AREADENSITYMEASURE = 'IfcAreaDensityMeasure'
AREAMEASURE = 'IfcAreaMeasure'
BINARY = 'IfcBinary'
BOOLEAN = 'IfcBoolean'
BOXALIGNMENT = 'IfcBoxAlignment'
CARDINALPOINTREFERENCE = 'IfcCardinalPointReference'
COMPLEXNUMBER = 'IfcComplexNumber'
COMPOUNDPLANEANGLEMEASURE = 'IfcCompoundPlaneAngleMeasure'
CONTEXTDEPENDENTMEASURE = 'IfcContextDependentMeasure'
COUNTMEASURE = 'IfcCountMeasure'
CURVATUREMEASURE = 'IfcCurvatureMeasure'
DATE = 'IfcDate'
DATETIME = 'IfcDateTime'
DAYINMONTHNUMBER = 'IfcDayInMonthNumber'
DAYINWEEKNUMBER = 'IfcDayInWeekNumber'
DESCRIPTIVEMEASURE = 'IfcDescriptiveMeasure'
DIMENSIONCOUNT = 'IfcDimensionCount'
DOSEEQUIVALENTMEASURE = 'IfcDoseEquivalentMeasure'
DURATION = 'IfcDuration'
DYNAMICVISCOSITYMEASURE = 'IfcDynamicViscosityMeasure'
ELECTRICCAPACITANCEMEASURE = 'IfcElectricCapacitanceMeasure'
ELECTRICCHARGEMEASURE = 'IfcElectricChargeMeasure'
ELECTRICCONDUCTANCEMEASURE = 'IfcElectricConductanceMeasure'
ELECTRICCURRENTMEASURE = 'IfcElectricCurrentMeasure'
ELECTRICRESISTANCEMEASURE = 'IfcElectricResistanceMeasure'
ELECTRICVOLTAGEMEASURE = 'IfcElectricVoltageMeasure'
ENERGYMEASURE = 'IfcEnergyMeasure'
FONTSTYLE = 'IfcFontStyle'
FONTVARIANT = 'IfcFontVariant'
FONTWEIGHT = 'IfcFontWeight'
FORCEMEASURE = 'IfcForceMeasure'
FREQUENCYMEASURE = 'IfcFrequencyMeasure'
GLOBALLYUNIQUEID = 'IfcGloballyUniqueId'
HEATFLUXDENSITYMEASURE = 'IfcHeatFluxDensityMeasure'
HEATINGVALUEMEASURE = 'IfcHeatingValueMeasure'
IDENTIFIER = 'IfcIdentifier'
ILLUMINANCEMEASURE = 'IfcIlluminanceMeasure'
INDUCTANCEMEASURE = 'IfcInductanceMeasure'
INTEGER = 'IfcInteger'
INTEGERCOUNTRATEMEASURE = 'IfcIntegerCountRateMeasure'
IONCONCENTRATIONMEASURE = 'IfcIonConcentrationMeasure'
ISOTHERMALMOISTURECAPACITYMEASURE = 'IfcIsothermalMoistureCapacityMeasure'
KINEMATICVISCOSITYMEASURE = 'IfcKinematicViscosityMeasure'
LABEL = 'IfcLabel'
LANGUAGEID = 'IfcLanguageId'
LENGTHMEASURE = 'IfcLengthMeasure'
LINEARFORCEMEASURE = 'IfcLinearForceMeasure'
LINEARMOMENTMEASURE = 'IfcLinearMomentMeasure'
LINEARSTIFFNESSMEASURE = 'IfcLinearStiffnessMeasure'
LINEARVELOCITYMEASURE = 'IfcLinearVelocityMeasure'
LINEINDEX = 'IfcLineIndex'
LOGICAL = 'IfcLogical'
LUMINOUSFLUXMEASURE = 'IfcLuminousFluxMeasure'
LUMINOUSINTENSITYDISTRIBUTIONMEASURE = 'IfcLuminousIntensityDistributionMeasure'
LUMINOUSINTENSITYMEASURE = 'IfcLuminousIntensityMeasure'
MAGNETICFLUXDENSITYMEASURE = 'IfcMagneticFluxDensityMeasure'
MAGNETICFLUXMEASURE = 'IfcMagneticFluxMeasure'
MASSDENSITYMEASURE = 'IfcMassDensityMeasure'
MASSFLOWRATEMEASURE = 'IfcMassFlowRateMeasure'
MASSMEASURE = 'IfcMassMeasure'
MASSPERLENGTHMEASURE = 'IfcMassPerLengthMeasure'
MODULUSOFELASTICITYMEASURE = 'IfcModulusOfElasticityMeasure'
MODULUSOFLINEARSUBGRADEREACTIONMEASURE = 'IfcModulusOfLinearSubgradeReactionMeasure'
MODULUSOFROTATIONALSUBGRADEREACTIONMEASURE = 'IfcModulusOfRotationalSubgradeReactionMeasure'
MODULUSOFSUBGRADEREACTIONMEASURE = 'IfcModulusOfSubgradeReactionMeasure'
MOISTUREDIFFUSIVITYMEASURE = 'IfcMoistureDiffusivityMeasure'
MOLECULARWEIGHTMEASURE = 'IfcMolecularWeightMeasure'
MOMENTOFINERTIAMEASURE = 'IfcMomentOfInertiaMeasure'
MONETARYMEASURE = 'IfcMonetaryMeasure'
MONTHINYEARNUMBER = 'IfcMonthInYearNumber'
NONNEGATIVELENGTHMEASURE = 'IfcNonNegativeLengthMeasure'
NORMALISEDRATIOMEASURE = 'IfcNormalisedRatioMeasure'
NUMERICMEASURE = 'IfcNumericMeasure'
PARAMETERVALUE = 'IfcParameterValue'
PHMEASURE = 'IfcPHMeasure'
PLANARFORCEMEASURE = 'IfcPlanarForceMeasure'
PLANEANGLEMEASURE = 'IfcPlaneAngleMeasure'
POSITIVEINTEGER = 'IfcPositiveInteger'
POSITIVELENGTHMEASURE = 'IfcPositiveLengthMeasure'
POSITIVEPLANEANGLEMEASURE = 'IfcPositivePlaneAngleMeasure'
POSITIVERATIOMEASURE = 'IfcPositiveRatioMeasure'
POWERMEASURE = 'IfcPowerMeasure'
PRESENTABLETEXT = 'IfcPresentableText'
PRESSUREMEASURE = 'IfcPressureMeasure'
PROPERTYSETDEFINITIONSET = 'IfcPropertySetDefinitionSet'
RADIOACTIVITYMEASURE = 'IfcRadioActivityMeasure'
RATIOMEASURE = 'IfcRatioMeasure'
REAL = 'IfcReal'
ROTATIONALFREQUENCYMEASURE = 'IfcRotationalFrequencyMeasure'
ROTATIONALMASSMEASURE = 'IfcRotationalMassMeasure'
ROTATIONALSTIFFNESSMEASURE = 'IfcRotationalStiffnessMeasure'
SECTIONALAREAINTEGRALMEASURE = 'IfcSectionalAreaIntegralMeasure'
SECTIONMODULUSMEASURE = 'IfcSectionModulusMeasure'
SHEARMODULUSMEASURE = 'IfcShearModulusMeasure'
SOLIDANGLEMEASURE = 'IfcSolidAngleMeasure'
SOUNDPOWERLEVELMEASURE = 'IfcSoundPowerLevelMeasure'
SOUNDPOWERMEASURE = 'IfcSoundPowerMeasure'
SOUNDPRESSURELEVELMEASURE = 'IfcSoundPressureLevelMeasure'
SOUNDPRESSUREMEASURE = 'IfcSoundPressureMeasure'
SPECIFICHEATCAPACITYMEASURE = 'IfcSpecificHeatCapacityMeasure'
SPECULAREXPONENT = 'IfcSpecularExponent'
SPECULARROUGHNESS = 'IfcSpecularRoughness'
TEMPERATUREGRADIENTMEASURE = 'IfcTemperatureGradientMeasure'
TEMPERATURERATEOFCHANGEMEASURE = 'IfcTemperatureRateOfChangeMeasure'
TEXT = 'IfcText'
TEXTALIGNMENT = 'IfcTextAlignment'
TEXTDECORATION = 'IfcTextDecoration'
TEXTFONTNAME = 'IfcTextFontName'
TEXTTRANSFORMATION = 'IfcTextTransformation'
THERMALADMITTANCEMEASURE = 'IfcThermalAdmittanceMeasure'
THERMALCONDUCTIVITYMEASURE = 'IfcThermalConductivityMeasure'
THERMALEXPANSIONCOEFFICIENTMEASURE = 'IfcThermalExpansionCoefficientMeasure'
THERMALRESISTANCEMEASURE = 'IfcThermalResistanceMeasure'
THERMALTRANSMITTANCEMEASURE = 'IfcThermalTransmittanceMeasure'
THERMODYNAMICTEMPERATUREMEASURE = 'IfcThermodynamicTemperatureMeasure'
TIME = 'IfcTime'
TIMEMEASURE = 'IfcTimeMeasure'
TIMESTAMP = 'IfcTimeStamp'
TORQUEMEASURE = 'IfcTorqueMeasure'
URIREFERENCE = 'IfcURIReference'
VAPORPERMEABILITYMEASURE = 'IfcVaporPermeabilityMeasure'
VOLUMEMEASURE = 'IfcVolumeMeasure'
VOLUMETRICFLOWRATEMEASURE = 'IfcVolumetricFlowRateMeasure'
WARPINGCONSTANTMEASURE = 'IfcWarpingConstantMeasure'
WARPINGMOMENTMEASURE = 'IfcWarpingMomentMeasure'


from __future__ import annotations
import SOMcreator
from SOMcreator import classes
from SOMcreator.filehandling import core
from SOMcreator.filehandling.constants import OBJECT, CONNECTION, AGGREGATIONS, PARENT
from SOMcreator.filehandling.typing import AggregationDict

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from SOMcreator import Project
    from SOMcreator.filehandling.typing import MainDict


### Import ###
def _get_aggregation(proj, aggregation_dict: dict, identifier: str, ):
    name, description, optional, parent, filter_matrix = core.get_basics(proj, aggregation_dict)
    object_uuid = aggregation_dict[OBJECT]
    obj = classes.get_element_by_uuid(object_uuid)
    parent_connection = aggregation_dict[CONNECTION]
    aggregation = classes.Aggregation(obj=obj, parent_connection=parent_connection, uuid=identifier,
                                      description=description, optional=optional, filter_matrix=filter_matrix)
    SOMcreator.filehandling.aggregation_dict[aggregation] = (parent, parent_connection)


def load(proj: classes.Project, main_dict: dict):
    aggregations_dict: dict[str, AggregationDict] = main_dict.get(AGGREGATIONS)
    core.remove_part_of_dict(AGGREGATIONS)
    aggregations_dict = dict() if core.check_dict(aggregations_dict, AGGREGATIONS) else aggregations_dict
    for uuid_ident, entity_dict in aggregations_dict.items():
        _get_aggregation(proj, entity_dict, uuid_ident)


def calculate():
    for aggregation, (uuid, connection_type) in SOMcreator.filehandling.aggregation_dict.items():
        parent = classes.get_element_by_uuid(uuid)
        if parent is None:
            continue
        parent.add_child(aggregation, connection_type)


### Export ###
def _create_entry(element: classes.Aggregation) -> AggregationDict:
    aggregation_dict: AggregationDict = dict()
    core.write_basics(aggregation_dict, element)
    aggregation_dict[OBJECT] = element.object.uuid
    if element.parent is not None:
        aggregation_dict[PARENT] = element.parent.uuid
    else:
        aggregation_dict[PARENT] = str(element.parent)
    aggregation_dict[CONNECTION] = element.parent_connection
    return aggregation_dict


def write(proj: Project, main_dict: MainDict):
    main_dict[AGGREGATIONS] = dict()
    for aggregation in proj.get_all_aggregations():
        main_dict[AGGREGATIONS][aggregation.uuid] = _create_entry(aggregation)



from .ifc_datatypes import LABEL, REAL, BOOLEAN, DATE, INTEGER

AGGREGATION = 1
INHERITANCE = 2

VALUE = "Value"
FORMAT = "Format"
RANGE = "Range"
LIST = "List"
SHARED_PARAMETERS = "SharedParameters"
EXISTS = "Exists"
VALUE_TYPE_LOOKUP = { FORMAT: "Format", RANGE: "Range", LIST: "List", } #VALUE: "Value",
RANGE_STRINGS = ["Range", "range", "RANGE"]

# DATATYPES
DATA_TYPES = [LABEL, INTEGER, REAL, BOOLEAN, DATE]
NUMBER_DATATYPES = [REAL, INTEGER]

# Datatype Mappings
OLD_DATATYPE_DICT = {"xs:string": LABEL,
                     "xs:double": REAL,
                     "xs:boolean": BOOLEAN,
                     "xs:long": REAL,
                     "xs:int": INTEGER}
XS_DATATYPE_DICT = {
    LABEL: "xs:string",
    REAL: "xs:double",
    BOOLEAN: "xs:boolean",
    INTEGER: "xs:int"
}
REVIT_TEMPLATE_DATATYPE_DICT = {INTEGER: "Integer",
                                LABEL: "Label",
                                REAL: "Real",
                                BOOLEAN: "Boolean",
                                DATE: "Date"}

REVIT_SHARED_PARAM_DATATYPE_DICT = {INTEGER: "INTEGER",
                                    LABEL: "TEXT",
                                    REAL: "NUMBER",
                                    BOOLEAN: "YESNO",
                                    DATE: "DATE"}
DATATYPE_DICT = {
    LABEL: str,
    BOOLEAN: bool,
    INTEGER: int,
    REAL: float
}



from __future__ import annotations
from typing import TypedDict
class StandardDict(TypedDict):
    name: str
    optional: bool
    filter_matrix: list[list[bool]]
    parent: str | None
    description: str


class MainDict(TypedDict):
    Project: ProjectDict
    PredefinedPropertySets: dict[str, PropertySetDict]
    Objects: dict[str, ObjectDict]
    Aggregations: dict[str, AggregationDict]
    AggregationScenes: dict[str, AggregationScene]


class ProjectDict(TypedDict):
    name: str
    author: str
    version: str
    AggregationAttributeName: str
    AggregationPsetName: str
    current_project_phase: int
    current_use_case: int
    ProjectPhases: list[FilterDict]
    UseCases: list[FilterDict]
    filter_matrix: list[list[bool]]


class ObjectDict(StandardDict):
    IfcMappings: list[str]
    abbreviation: str
    ident_attribute: str
    PropertySets: dict[str, PropertySetDict]


class PropertySetDict(StandardDict):
    Attributes: dict[str, AttributeDict]


class AttributeDict(StandardDict):
    data_type: str
    value_type: str
    child_inherits_value: bool
    revit_mapping: str
    Value: list[str] | list[float] | list[[float, float]]


class AggregationDict(StandardDict):
    Object: str | None
    connection: int
    x_pos: float
    y_pos: float


class AggregationScene(TypedDict):
    Nodes: list[str]


class FilterDict(TypedDict):
    name: str
    long_name: str
    description: str